/*
isr4pi.c
D. Thiebaut
based on isr.c from the WiringPi library, authored by Gordon Henderson
https://github.com/WiringPi/WiringPi/blob/master/examples/isr.c

Compile as follows:
    gcc -o isr4pi isr4pi.c -lwiringPi

Run as follows:
    sudo ./isr4pi
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <wiringPi.h>

// Use GPIO Pin..., which is Pin... for wiringPi library

#define INTERRUPT_PIN 2
#define SPI_CHAN 0
#define BASE 100

#define HIGH 1
#define LOW 0
#define LED_COLUMNS 8
#define LED_ROWS 8
#define PHOTOTRANSISTOR_COLUMNS 8
#define PHOTOTRANSISTOR_ROWS 8

#define MUX_S0 1
#define MUX_S1 4
#define MUX_S2 5

#define INT_MUX_S0 23
#define INT_MUX_S1 22
#define INT_MUX_S2 21

#define BITMASK 0x1
#define SRCLR_  29
#define SRCK  28
#define RCK 27
#define SERIN 8
#define G_ 9

struct wiringPiNodeStruct *node ;

int myAnalogRead (int pin)
{
	unsigned char spiData [3] ;
	unsigned char chanBits ;
	int chan = pin - node->pinBase ;
	
	chanBits = 0b10000000 | (chan << 4) ;

	spiData [0] = 1 ;             // Start bit
	spiData [1] = chanBits ;
	spiData [2] = 0 ;

	wiringPiSPIDataRW (node->fd, spiData, 3) ;

	return ((spiData [1] << 8) | spiData [2]) & 0x3FF ;
}

int mcp3004Setup (const int pinBase, int spiChannel)
{
	if (wiringPiSPISetup (spiChannel, 1000000) < 0)
		return FALSE ;

	node = wiringPiNewNode (pinBase, 8) ;

	node->fd         = spiChannel ;
	node->analogRead = myAnalogRead ;

	return TRUE ;
}

// the event counter
volatile int eventCounter = 0;
volatile int grid[PHOTOTRANSISTOR_ROWS][PHOTOTRANSISTOR_COLUMNS];

// -------------------------------------------------------------------------
// myInterrupt:  called every time an event occurs
void myInterrupt(void)
{
	eventCounter++;

	int i;
	struct timeval stop, start, start_full, stop_full;

	int s0, s1, s2;

	gettimeofday(&start_full, NULL);

	for (s2 = 0; s2 <= 1; s2++)
	{
		for (s1 = 0; s1 <= 1; s1++)
		{
			for (s0 = 0; s0 <= 1; s0++)
			{
				digitalWrite(MUX_S0, s0);
				digitalWrite(MUX_S1, s1);
				digitalWrite(MUX_S2, s2);
				int row = (s2 << 2) | (s1 << 1) | s0;

				for (i = 0; i < PHOTOTRANSISTOR_COLUMNS; i++)
				{
					//gettimeofday(&start, NULL);
					grid[row][i] = myAnalogRead(BASE+i);
                    //gettimeofday(&stop, NULL);
                    //if (stop.tv_usec-start.tv_usec > 600) printf("FUCKKKKKK %lu  \n",stop.tv_usec-start.tv_usec);
				}
			}
		}
	}
	gettimeofday(&stop_full, NULL);
	//printf("start %lu \n", start_full.tv_usec);
	//printf("stop %lu \n", stop_full.tv_usec);
	//printf("total array time: %lu\n", stop_full.tv_usec-start_full.tv_usec);
	int x;
	for (x = 0; x < 8; x++) {
		printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", grid[x][0],grid[x][1],grid[x][2],grid[x][3],grid[x][4],grid[x][5],grid[x][6],grid[x][7]);
	}
	printf("\n");
	//give this thread the highest priority possible
	delay(0);
}

// -------------------------------------------------------------------------
// main
int main(void)
{
	// sets up the wiringPi library
	if (wiringPiSetup () < 0) {
		fprintf (stderr, "Unable to setup wiringPi: %s\n", strerror (errno));
		return 1;
	}

	mcp3004Setup(BASE,SPI_CHAN);
	//test shift registers
	pinMode(SRCLR_, OUTPUT);
	pinMode(SRCK, OUTPUT);
	pinMode(RCK, OUTPUT);
	pinMode(SERIN, OUTPUT);
	pinMode(G_, OUTPUT);

	digitalWrite(SRCK, LOW);
	digitalWrite(RCK, LOW); // start clks low

	digitalWrite(G_, LOW); // active low, show register to output

	digitalWrite(SRCLR_, LOW); // active low, clear all shift registers
	digitalWrite(SRCLR_, HIGH); // asynchronous?

	char test[LED_ROWS];
	//memset(test, '\0', 8);
	memset(test, 0x0F, 1);
	memset(test+1,0x0F,1);
	memset(test+2,0x0F,1);
	memset(test+3,0x0F,1);
	memset(test+4,0x00,1);
	memset(test+5,0x00,1);
	memset(test+6,0x00,1);
	memset(test+7,0x00,1);
	//memset(test, 0x55, 1);

	printf("test %c %c %c %c %c %c %c %c\n", test[0], test[1], test[2], test[3], test[4], test[5], test[6], test[7]);

	int a, b;
	for (a = 0; a < LED_ROWS; a++)
	{
		for (b = 0; b < LED_COLUMNS; b++)
		{
			digitalWrite(SERIN, (test[a] >> b) & BITMASK);
			//delay(1000);
			digitalWrite(SRCK, HIGH);
			//delay(1000);
			digitalWrite(SRCK, LOW);
		}
	}

	digitalWrite(RCK, HIGH);
	//delay(1000);
	digitalWrite(RCK, LOW); // set output registers with serial registers

	pinMode(MUX_S0, OUTPUT);
	pinMode(MUX_S1, OUTPUT);
	pinMode(MUX_S2, OUTPUT);
	while(1) {
		myInterrupt();
		delay(5000);
	}
	while(1);
	//sample array
	pinMode(INTERRUPT_PIN, INPUT);

	pinMode(INT_MUX_S1, OUTPUT);
	pinMode(INT_MUX_S0, OUTPUT);
	pinMode(INT_MUX_S2, OUTPUT);
	
	//select channel 2
	digitalWrite(INT_MUX_S0, LOW);
	digitalWrite(INT_MUX_S1, HIGH);
	digitalWrite(INT_MUX_S2, LOW);

	//generate an interrupt on high-to-low transitions
	//and attach myInterrupt() to the interrupt
	if ( wiringPiISR (INTERRUPT_PIN, INT_EDGE_RISING, &myInterrupt) < 0 ) {
		fprintf (stderr, "Unable to setup ISR: %s\n", strerror (errno));
		return 1;
	}

	//delay( 5000 );

	//display counter value every second.
	while ( 1 )
	{
		//select channel 1
		digitalWrite(INT_MUX_S0, LOW);
		digitalWrite(INT_MUX_S1, LOW);
		digitalWrite(INT_MUX_S2, HIGH);

		//delay( 1000 ); // wait 1 second

		//select channel 2
		digitalWrite(INT_MUX_S0, LOW);
		digitalWrite(INT_MUX_S1, HIGH);
		digitalWrite(INT_MUX_S2, LOW);

		//delay( 3000);

		//printf( "count %d\n", eventCounter );
	}
}
