/*
isr4pi.c
D. Thiebaut
based on isr.c from the WiringPi library, authored by Gordon Henderson
https://github.com/WiringPi/WiringPi/blob/master/examples/isr.c

Compile as follows:
    gcc -o isr4pi isr4pi.c -lwiringPi

Run as follows:
    sudo ./isr4pi
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <wiringPi.h>
#include <limits.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <math.h>

#include <arpa/inet.h>

#define PORT "3490"
#define MAXDATASIZE 100

// Use GPIO Pin..., which is Pin... for wiringPi library

#define TEST_ON_PI
#define PRINT_SAMPLES
#define PRINT_CONNECTED_COMPONENTS
//#define TIME_REQUIRED_SAMPLE
//#define ROW_INDEX
//#define TEST_WIFI
//#define PRINT_THRESHOLD
//#define OFFSET

#define TALLEST_PERSON 6
#define MAX_GROUPS 10
#define IP_ADDRESS "192.168.43.123\0"
//#define IP_ADDRESS "192.168.2.35\0"
//#define IP_ADDRESS "127.0.0.1\0"
#define MINIMUM_PIXELS 6
#define ADC_STEP 4.88
#define MAX_CURRENT 0.020
#define MAX_PATTERN 5
#define THRESHOLD_VALUE 13
#define CALIBRATION_TIME 2
#define RESISTOR 22.4 * pow(10,6)
#define HEIGHT 8.0
#define MAX_VOLTAGE 4.8

#define INTERRUPT_PIN 2
#define SPI_CHAN 0
#define BASE 100

#define HIGH 1
#define LOW 0
#define LED_COLUMNS 8
#define LED_ROWS 8
#define PHOTOTRANSISTOR_COLUMNS 8
#define PHOTOTRANSISTOR_ROWS 8

#define MUX_S0 1
#define MUX_S1 4
#define MUX_S2 5

#define INT_MUX_S0 23
#define INT_MUX_S1 22
#define INT_MUX_S2 21

#define BITMASK 0x1
#define SRCLR_  29
#define SRCK  28
#define RCK 27
#define SERIN 8
#define G_ 9

#define NROWS 8
#define NCOLS 8
#define NPIXELS 64
#define MAX_INTENSITY 1024
#define ERROR -1

//calibration flag
int initial_calibration = 1;

// group count
int previous_number_groups = -1;

// time variable
double  highest_cpu_time = 0.0;


// server variables

char buf[MAXDATASIZE];
int sockfd;

// raster scan variables

char histogram[MAX_INTENSITY+1];
char connected_components[NROWS][NCOLS];
struct wiringPiNodeStruct *node ;

// the event counter

int grid[PHOTOTRANSISTOR_ROWS][PHOTOTRANSISTOR_COLUMNS];
int test_grid[PHOTOTRANSISTOR_ROWS][PHOTOTRANSISTOR_COLUMNS] = {{29,30,31,12,27,28,14,15},{8,9,10,11,12,13,14,15},{16,17,18,19,20,21,22,23},{24,25,26,27,28,29,30,31},{0,34,2,3,4,5,6,7},{8,52,10,11,12,13,14,15},{16,17,18,19,20,21,22,23},{24,25,26,27,28,29,30,31}};
int offset[PHOTOTRANSISTOR_ROWS][PHOTOTRANSISTOR_COLUMNS] = {{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0},{0,0,0,0,0,0,0,0}};


// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int init_client(char * ip)
{
    struct addrinfo hints, *servinfo, *p;
    int rv;
    char s[INET6_ADDRSTRLEN];

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if ((rv = getaddrinfo(ip, PORT, &hints, &servinfo)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
        return 1;
    }

    // loop through all the results and connect to the first we can
    for(p = servinfo; p != NULL; p = p->ai_next) {
        if ((sockfd = socket(p->ai_family, p->ai_socktype,
                p->ai_protocol)) == -1) {
            perror("client: socket");
            continue;
        }

        if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
            close(sockfd);
            perror("client: connect");
            continue;
        }

        break;
    }

    if (p == NULL) {
        fprintf(stderr, "client: failed to connect\n");
        return 2;
    }

    inet_ntop(p->ai_family, get_in_addr((struct sockaddr *)p->ai_addr),
            s, sizeof s);
    printf("client: connecting to %s\n", s);

    freeaddrinfo(servinfo); // all done with this structure
}

int myAnalogRead (int pin)
{
	unsigned char spiData [3] ;
	unsigned char chanBits ;
	int chan = pin - node->pinBase ;

	chanBits = 0b10000000 | (chan << 4) ;

	spiData [0] = 1 ;             // Start bit
	spiData [1] = chanBits ;
	spiData [2] = 0 ;

	wiringPiSPIDataRW (node->fd, spiData, 3) ;

	return ((spiData [1] << 8) | spiData [2]) & 0x3FF ;
};

int mcp3004Setup (const int pinBase, int spiChannel)
{
	if (wiringPiSPISetup (spiChannel, 1000000) < 0)
		return FALSE ;

	node = wiringPiNewNode (pinBase, 8) ;

	node->fd         = spiChannel ;
	node->analogRead = myAnalogRead ;

	return TRUE ;
};

// -------------------------------------------------------------------------
// myInterrupt:  called every time an event occurs



int sample_array(int set_offset_array)
{

	int i;
	struct timeval stop, start, start_full, stop_full;

	int s0, s1, s2;
	clock_t start_time, stop_time;
	double cpu_time_used;

	#ifdef TEST_ON_PI
	//gettimeofday(&start_full, NULL);
	start_time = clock();

	for (s2 = 0; s2 <= 1; s2++)
	{
		for (s1 = 0; s1 <= 1; s1++)
		{
			for (s0 = 0; s0 <= 1; s0++)
			{
				digitalWrite(MUX_S0, s0); // select column
				digitalWrite(MUX_S1, s1);
				digitalWrite(MUX_S2, s2);
				int row = (s2 << 2) | (s1 << 1) | s0;
				#ifdef ROW_INDEX
				printf("row selected is %d\n", row);
				#endif
				if (set_offset_array) {
					for (i = 0; i < PHOTOTRANSISTOR_COLUMNS; i++)
					{
						offset[row][i] = myAnalogRead(BASE+i); // select row
						//printf("filling offset array\n");

					}
				}
				else {
					for (i = 0; i < PHOTOTRANSISTOR_COLUMNS; i++)
					{
						//gettimeofday(&start, NULL);
						int value = myAnalogRead(BASE+i) - offset[row][i]; // select row
	                    			if (value > 0) {
							test_grid[row][i] = value;
						}
						else {
							test_grid[row][i] = 0;
						}
				//gettimeofday(&stop, NULL);
	                    //if (stop.tv_usec-start.tv_usec > 600) printf("over .6 ms per sample! %lu  \n",stop.tv_usec-start.tv_usec);
					}
				}
				#ifdef ROW_INDEX
					printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", test_grid[row][0],test_grid[row][1],test_grid[row][2],test_grid[row][3],test_grid[row][4],test_grid[row][5],test_grid[row][6],test_grid[row][7]);
					printf("\n");
					delay(1000);
				#endif
			}
		}
	}

	//gettimeofday(&stop_full, NULL);
	stop_time = clock();
	cpu_time_used = ((double) (stop_time-start_time)) / CLOCKS_PER_SEC;
	#ifdef TIME_REQUIRED_SAMPLE
	if (cpu_time_used > highest_cpu_time) {
		highest_cpu_time = cpu_time_used;
		printf("total array time: %lf seconds\n", cpu_time_used);
	}
	#endif

	#endif 


	#ifdef PRINT_SAMPLES
	int x;
	for (x = 0; x < 8; x++) {
	printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", test_grid[x][0],test_grid[x][1],test_grid[x][2],test_grid[x][3],test_grid[x][4],test_grid[x][5],test_grid[x][6],test_grid[x][7]);
	}
	printf("\n");
	delay(4000);
	#endif

	//give this thread the highest priority possible
	//delay(0);
};

void test_shift_registers(int bytes_on) {
	//test shift registers
	pinMode(SRCLR_, OUTPUT);
	pinMode(SRCK, OUTPUT);
	pinMode(RCK, OUTPUT);
	pinMode(SERIN, OUTPUT);
	pinMode(G_, OUTPUT);

	digitalWrite(SRCK, LOW);
	digitalWrite(RCK, LOW); // start clks low

	digitalWrite(G_, LOW); // active low, show register to output

	digitalWrite(SRCLR_, LOW); // active low, clear all shift registers
	digitalWrite(SRCLR_, HIGH); // asynchronous?

	char test[LED_ROWS];
	memset(test, 0x00, LED_ROWS);
	memset(test, 0xFF, bytes_on);
	//memset(test, 0xFF, 8);
	//memset(test, 0x55, 1);

	//printf("test %c %c %c %c %c %c %c %c\n", test[0], test[1], test[2], test[3], test[4], test[5], test[6], test[7]);

	int a, b;
	for (a = 0; a < LED_ROWS; a++)
	{
		for (b = 0; b < LED_COLUMNS; b++)
		{
			digitalWrite(SERIN, (test[a] >> b) & BITMASK);
			//delay(1000);
			digitalWrite(SRCK, HIGH);
			//delay(1000);
			digitalWrite(SRCK, LOW);
		}
	}

	digitalWrite(RCK, HIGH);
	//delay(1000);
	digitalWrite(RCK, LOW); // set output registers with serial registers

	//printf("test shift registers\n");
};

// void test_interrupt_mux() {

// 	pinMode(INTERRUPT_PIN, INPUT);

// 	pinMode(INT_MUX_S1, OUTPUT);
// 	pinMode(INT_MUX_S0, OUTPUT);
// 	pinMode(INT_MUX_S2, OUTPUT);
	
// 	//select channel 2
// 	digitalWrite(INT_MUX_S0, LOW);
// 	digitalWrite(INT_MUX_S1, HIGH);
// 	digitalWrite(INT_MUX_S2, LOW);

// 	//generate an interrupt on high-to-low transitions
// 	//and attach myInterrupt() to the interrupt
// 	if ( wiringPiISR (INTERRUPT_PIN, INT_EDGE_RISING, &myInterrupt) < 0 ) {
// 		fprintf (stderr, "Unable to setup ISR: %s\n", strerror (errno));
// 		return;
// 	}

// 	//delay( 5000 );

// 	//display counter value every second.
// 	while ( 1 )
// 	{
// 		//select channel 1
// 		digitalWrite(INT_MUX_S0, LOW);
// 		digitalWrite(INT_MUX_S1, LOW);
// 		digitalWrite(INT_MUX_S2, HIGH);

// 		//delay( 1000 ); // wait 1 second

// 		//select channel 2
// 		digitalWrite(INT_MUX_S0, LOW);
// 		digitalWrite(INT_MUX_S1, HIGH);
// 		digitalWrite(INT_MUX_S2, LOW);

// 		//delay( 3000);

// 		//printf( "count %d\n", eventCounter );
// 	}

// };

int number_of_groups() {

	int x, y;
	char groups[MAX_GROUPS];
	memset(groups, 0, MAX_GROUPS);

	for (x = 0; x < PHOTOTRANSISTOR_COLUMNS; x++) {
		for (y=0; y < PHOTOTRANSISTOR_ROWS; y++) {
			int current_group = connected_components[y][x];
			if (current_group > 0 ) 
				//printf("current group is %d\n", current_group);
				groups[current_group] += 1;
		}
	}

	//printf("groups %d %d %d %d %d %d %d %d %d %d\n", groups[0], groups[1], groups[2], groups[3], groups[4], groups[5], groups[6], groups[7], groups[8], groups[9]);

	int i;
	int count = 0;
	for (i = 1; i < MAX_GROUPS; i++) {
		if (groups[i] >= MINIMUM_PIXELS) count += 1;
	}

	return count;

}

int fill_hist( int intensity ) {
//	printf("intensity %d\n", intensity);
	histogram[intensity]++;
};


float P( int z, int num_rows, int num_cols ) {
	
	return (float)histogram[z] / (num_rows * num_cols);
};

float u0 ( int z, float p, float q0_prev, float q0, float u0_prev ) {
//	printf("part 1 is %f and part 2 is %f\n", (z * p)/q0, (q0_prev/q0)*u0_prev);
	if (q0 == 0) return 0;
	return (( z * p ) / q0) + (q0_prev / q0 )*u0_prev;
};

float u1 ( float z, float u, float q0, float u0) {
	if (1-q0 == 0) return 0;
	return ( u - (q0 * u0) ) / ( 1 - q0);
};

float u (int num_rows, int num_cols) {
	
	float sum = 0;
	int i;
	for (i = 0; i <= MAX_INTENSITY; i++) {
		sum += i * P(i, num_rows, num_cols);
	}
	return sum;
};

float q0 (float probability, int z, float q0_prev ) {
	return probability + q0_prev;

};

int clear_and_populate_histogram(short * array_to_read, int num_short) {

	memset((char*)histogram, '\0', 1024); // clear histogram

	int i;
	for (i = 0; i < num_short; i++) {
		fill_hist(array_to_read[i]);
	}

};

int find_threshold(int num_rows, int num_cols) {

	//clock_t start, end;
	//double cpu_time_used;
	//start = clock();

	int zt;
	float u0_prev = 0;
	float q0_prev = 0;
	float mu = u(num_rows, num_cols);
	float sigma_squared = INT_MIN;
	int zt_chosen = 0;


	for (zt = 0 ; zt <= MAX_INTENSITY ; zt++) {
		//printf("zt is %d\n", zt);
		float probability = P (zt, num_rows, num_cols);
		//printf("probability is %f\n", probability);
		float q_naught = q0 (probability, zt, q0_prev);
		//printf("q_naught is %f\n", q_naught);
		float u_naught = u0( zt, probability, q0_prev, q_naught, u0_prev);
		//printf("u_naught is %f\n", u_naught);
		float u_one = u1( zt, mu, q_naught, u_naught);
		//printf("u_one is %f\n", u_one);
		float sigma_squared_new = q_naught * (1 - q_naught) * (u_naught - u_one) * (u_naught - u_one);
		if (sigma_squared_new > sigma_squared ) {
			sigma_squared = sigma_squared_new;
			zt_chosen = zt;
		}

		q0_prev = q_naught;
		u0_prev = u_naught;


	}
	// end = clock();

	// cpu_time_used = ((double) (end-start)) / CLOCKS_PER_SEC;
	int a;
	//for (a = 0; a <= MAX_INTENSITY; a++) {
	//	printf("h[%d]: %d\n", a, histogram[a]);
	//}

	//printf("threshold chosen is %d\n", zt_chosen);
	// printf("time used is %f\n", cpu_time_used);
	return zt_chosen;

};

first_raster_scan( int zt_chosen, short * image, int num_cols, int num_rows ) {

	memset(connected_components, '\0', NPIXELS);
	int min_group_num = 1;
	int y, x;
	for (y = 0; y < num_rows; y++) 
	{
		for (x = 0; x < num_cols; x++) 
		{
			//printf("got there\n");

			// for (y=0; y<8; y++) 
			// {
			// 	printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", connected_components[y][0],connected_components[y][1],connected_components[y][2],connected_components[y][3],connected_components[y][4],connected_components[y][5],connected_components[y][6],connected_components[y][7]);
			// }

			if (test_grid[y][x] >= zt_chosen) 
			{
				// check above
				if ( y - 1 >= 0 )
				{
					int group_number = connected_components[y-1][x];
					//printf("group number is %d\n", group_number);
					if ((group_number < min_group_num) && group_number > 0)
						min_group_num  = group_number;
				}

				// check left
				if ( x - 1 >= 0 )
				{
					int group_number = connected_components[y][x-1];
					if ((group_number < min_group_num) && group_number > 0)
						min_group_num = group_number;
				}

				if ( min_group_num > 0 )
				{
					connected_components[y][x] = min_group_num;
					min_group_num++;
				}
			}
		}
	}

	// for (y=0; y<8; y++)
	// {
	// 	printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", connected_components[y][0],connected_components[y][1],connected_components[y][2],connected_components[y][3],connected_components[y][4],connected_components[y][5],connected_components[y][6],connected_components[y][7]);
	// }
	// printf("\n");

};

second_raster_scan(int num_cols, int num_rows) {

	int x, y;
	
	for (y = 0; y < num_rows; y++) {
		for (x = 0; x < num_cols; x++) {
			// we want to select the minimum group number of an adjacent neighbor
			int min_group_num = INT_MAX;
			if ( y-1 >= 0 && connected_components[y-1][x] > 0) {
				min_group_num = connected_components[y-1][x];
			}

			if ( y + 1 <= NROWS - 1) {
				int group_number = connected_components[y+1][x];
				if (group_number < min_group_num && group_number > 0)
					min_group_num = group_number;
			}

			if ( x - 1 >= 0 ) {
				int group_number = connected_components[y][x-1];
				if (group_number < min_group_num && group_number > 0)
					min_group_num = group_number;
			}

			if ( x + 1 <= NCOLS - 1) {
				int group_number = connected_components[y][x+1];
				if (group_number < min_group_num && group_number > 0)
					min_group_num = group_number;
			}

			if (min_group_num < connected_components[y][x]) {
				connected_components[y][x] = min_group_num;
			}

		}
	}

	#ifdef PRINT_CONNECTED_COMPONENTS
	for (y=0; y<8; y++) 
	{
		printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", connected_components[y][0],connected_components[y][1],connected_components[y][2],connected_components[y][3],connected_components[y][4],connected_components[y][5],connected_components[y][6],connected_components[y][7]);
	}
	printf("\n");
	delay(1000);
	#endif

};

void adc_select_setup() {

	// init wiringPi
	if (wiringPiSetup () < 0) {
		fprintf (stderr, "Unable to setup wiringPi: %s\n", strerror (errno));
		return;
	}

	mcp3004Setup(BASE,SPI_CHAN);

	//indices for choosing array column
	pinMode(MUX_S0, OUTPUT);
	pinMode(MUX_S1, OUTPUT);
	pinMode(MUX_S2, OUTPUT);
};

void sample_to_histogram() {
	int x, y;

	short perim[PHOTOTRANSISTOR_COLUMNS*PHOTOTRANSISTOR_ROWS];

	for (x = 0; x < PHOTOTRANSISTOR_COLUMNS; x++) {
		for (y = 0; y < PHOTOTRANSISTOR_ROWS; y++) {
			perim[y*PHOTOTRANSISTOR_COLUMNS + x] = test_grid[y][x];
		}
	}
	clear_and_populate_histogram(perim, PHOTOTRANSISTOR_COLUMNS*PHOTOTRANSISTOR_ROWS);
}

int threshold_full_array() {
	int x, y;

	short perim[PHOTOTRANSISTOR_COLUMNS*PHOTOTRANSISTOR_ROWS];

	for (x = 0; x < PHOTOTRANSISTOR_COLUMNS; x++) {
		for (y = 0; y < PHOTOTRANSISTOR_ROWS; y++) {
			perim[y*PHOTOTRANSISTOR_COLUMNS + x] = test_grid[y][x];
		}
	}

	clear_and_populate_histogram(perim, PHOTOTRANSISTOR_COLUMNS*PHOTOTRANSISTOR_ROWS); // now we take our big intensity array and get 1 threshold for all 4 sides
	int zt = find_threshold(1, PHOTOTRANSISTOR_COLUMNS*PHOTOTRANSISTOR_ROWS);

	// printf("found threshold, full_array\n");
	if (zt > THRESHOLD_VALUE) {
		first_raster_scan(zt, (short int *) *test_grid, 8, 8);
		second_raster_scan(8,8);
		second_raster_scan(8,8);
		second_raster_scan(8,8);
		#ifdef PRINT_THRESHOLD
		printf("threshold is %d\n", zt);
		#endif
		return 1; // found groups
	}
	else {
	//	printf("no groups found\n");
		return 0; // no groups found
	}
}



int sample_perimeter()
{
	short perim[4*PHOTOTRANSISTOR_COLUMNS - 4];
	short grid_temp[PHOTOTRANSISTOR_COLUMNS][PHOTOTRANSISTOR_ROWS];
	memset(grid_temp, '\0', 2*NPIXELS);

	int zt;

	int y = 0;
	int x = 0;

 // -------->

	#ifdef TEST_ON_PI
	// select row 0 
	digitalWrite(MUX_S0, 0);
	digitalWrite(MUX_S1, 0);
	digitalWrite(MUX_S2, 0);

	// sample all columns across row 0
	int i;
	for (i = 0; i < PHOTOTRANSISTOR_COLUMNS; i++)
	{
		perim[i] = myAnalogRead(BASE+i);
		grid_temp[0][i] = perim[i];
//		printf("%d,", perim[i]);
	}
//	printf("\n");
// ------->
// 		  |
// 		  |
// 		  |
// 		  v

	// select rows 0-7, column 7
	int row_idx = 7;
	int s0, s1, s2;
	for (s2 = 0; s2 <= 1; s2++)
	{
		for (s1 = 0; s1 <= 1; s1++)
		{
			for (s0 = 0; s0 <= 1; s0++)
			{
				digitalWrite(MUX_S0, s0);
				digitalWrite(MUX_S1, s1);
				digitalWrite(MUX_S2, s2);
					
				perim[row_idx] = myAnalogRead(BASE+(NCOLS-1));
				row_idx++;
			}
		}
	}
	// continue writing to big array of intensities containing all 4 sides
	for (i = 7; i < PHOTOTRANSISTOR_COLUMNS + 7; i++)
	{
//		printf("%d,", perim[i]);
		grid_temp[i-7][7] = perim[i];
	}
//	printf("\n");
	// there's redundant sampling where sides of perimeter meet

// ------->
// 		  |
// 		  |
// 		  |
// <----- v

	// select row 7
	digitalWrite(MUX_S0, 1);
	digitalWrite(MUX_S1, 1);
	digitalWrite(MUX_S2, 1);

	// sample all columns across row 7
	for (i = 14; i < PHOTOTRANSISTOR_COLUMNS+14; i++)
	{
		perim[i] = myAnalogRead(BASE+i);
//		printf("%d,", perim[i]);
		grid_temp[7][7 + 14 - i] = perim[i];
	}
//	printf("\n");

// ^  ------->
// |		  |
// |		  |
// |		  |
//   <-----   v

	// sample column 0, row 7 -> 0
	row_idx = 28;
	for (s2 = 1; s2 >= 0; s2--)
	{
		for (s1 = 1; s1 >= 0; s1--)
		{
			for (s0 = 1; s0 >= 0; s0--)
			{
				digitalWrite(MUX_S0, s0);
				digitalWrite(MUX_S1, s1);
				digitalWrite(MUX_S2, s2);
				perim[row_idx] = myAnalogRead(BASE);
				row_idx--;
			}
		}
	}

	perim[29] = perim[0];

	for (i = 21; i < PHOTOTRANSISTOR_COLUMNS+21; i++)
	{
//		printf("%d,", perim[i]);
		grid_temp[i-21][0] = perim[i];
	}

	#else

		// sample all columns across row 0
		

		for (x = 0; x < PHOTOTRANSISTOR_COLUMNS; x++)
		{
			perim[x + y] = test_grid[y][x];
		}

		// sample rows 0-7, column 7
		for ( y = 0; y < PHOTOTRANSISTOR_ROWS; y++ )
		{
			perim[x + y] = test_grid[y][ x + PHOTOTRANSISTOR_COLUMNS-1];
		}

		// y over bound
		y = y - 2;

		// sample all columns across row 7
		for ( ; x < 2*PHOTOTRANSISTOR_COLUMNS-1; x++)
		{
			perim[x + y] = test_grid[PHOTOTRANSISTOR_ROWS-1][2*PHOTOTRANSISTOR_COLUMNS -1 - x];
			printf("perim[x+y] is %d\n", perim[x+y]);
		}

		// x over bound
		x--;

		// sample column 0, row 7 -> 0
		for (y=PHOTOTRANSISTOR_ROWS-1 ; y < 2*PHOTOTRANSISTOR_ROWS-1; y++)
		{
			printf("%d\n", 2*PHOTOTRANSISTOR_ROWS-1-y);
			perim[x + y] = test_grid[2*(PHOTOTRANSISTOR_ROWS-1)-y][0];
		}

		int a, b;
		for (a = 0; a < 8; a++) {
			for (b = 0; b < 8; b++) {
				printf("%d,", test_grid[a][b]);
			}
			printf("\n");
		}


		int i = 0;
		for (; i < 28; i++) {
			printf("%d,", perim[i]);
		}
		printf("\n");

	#endif /*Row select bits on Pi */




//	printf("\n");
//	printf("ready to make hist\n");
	clear_and_populate_histogram(perim, 4*PHOTOTRANSISTOR_COLUMNS - 4); // now we take our big intensity array and get 1 threshold for all 4 sides
//	printf("histogram made\n");
	zt = find_threshold(1, 4*PHOTOTRANSISTOR_COLUMNS - 4);

	printf("found threshold, perimeter\n");
	if (zt > 15) {
		first_raster_scan(zt, (short int *) *test_grid, 8, 8);
		second_raster_scan(8,8);
		second_raster_scan(8,8);
		return 1; // found groups
	}
	else {
		//printf("no groups found\n");
		return 0; // no groups found
	}


	// printf("\n");
	// for (y=0; y<8; y++) 
	// {
	// 	printf("| %d | %d | %d | %d | %d | %d | %d | %d |\n", test_grid[y][0],test_grid[y][1],test_grid[y][2],test_grid[y][3],test_grid[y][4],test_grid[y][5],test_grid[y][6],test_grid[y][7]);
	// }


};




// -------------------------------------------------------------------------
// main
int main(void)
{

	// initialize client
	init_client(IP_ADDRESS);
	int numbytes;
	#ifdef TEST_ON_PI
	adc_select_setup();
	#endif
	double time_since_calibration = 0;

	while (1) {

		
		clock_t start_time, end_time;
		//sample_perimeter();

		start_time = clock();
		sample_array(0);
		if (threshold_full_array()) {
		//	memset(buf,'\0', MAXDATASIZE);
			//sprintf(buf, "we see %d groups", number_of_groups());
			//printf("%s\n", buf);
			//if ((numbytes = send(sockfd, buf, MAXDATASIZE-1, 0)) == -1) {
	      //      perror("recv");
	        //    exit(1);
	        //}
	   	//	buf[numbytes] = '\0';
			sample_array(0);
		}
		end_time = clock();
		double cpu_time = (float)(end_time - start_time)/CLOCKS_PER_SEC;
		// printf("time passed is %f\n", cpu_time);
		time_since_calibration += cpu_time;
		//printf("time since calibration %lf\n", time_since_calibration);
		// printf("time since calibration is %lf\n", time_since_calibration);
		if (initial_calibration || (time_since_calibration > CALIBRATION_TIME && previous_number_groups == 0)) {
			if (initial_calibration) {
				initial_calibration = 0;
			}
			int pattern_count = 0;
			int continue_calibration = 1;
			while (continue_calibration) {
				test_shift_registers(pattern_count); // turn off LEDs initially
				time_since_calibration = 0;
				sample_array(0);
				sample_to_histogram();
				int mu = u(PHOTOTRANSISTOR_ROWS, PHOTOTRANSISTOR_COLUMNS);
				float mean_voltage = ADC_STEP * mu * pow(10, -3);
				float mean_current = mean_voltage / RESISTOR;
				float expected_voltage = (mean_voltage * pow((HEIGHT/(HEIGHT - TALLEST_PERSON)),2)); // Vexpected(4 feet)>>Vu?
				float expected_current = expected_voltage / RESISTOR;
				// printf("mean_voltage is %lf\n", mean_voltage);
				// printf("mean_current is %lf\n", mean_current);
				// printf("expected voltage is %lf\n", expected_voltage);
				// printf("expected current is %lf\n", expected_current);
				// What should these conditions be?
				if ( expected_voltage < MAX_VOLTAGE &&  mean_current < MAX_CURRENT && expected_current < MAX_CURRENT && pattern_count < MAX_PATTERN) {
					pattern_count++;
				}
				else if (pattern_count > 0){ 
					test_shift_registers(pattern_count-1);
					continue_calibration = 0;
					#ifdef OFFSET
					sample_array(1);
					#endif
				//	printf("pattern count chosen is %d\n", pattern_count);
					delay(2000);

				}
				else {
					//printf("error!\n");
					continue_calibration = 0;
					#ifdef OFFSET
					sample_array(1);
					#endif
				}
				delay(2000);
			}
		}

		int number_groups = number_of_groups();

		// receive / send here
		memset(buf,'\0', MAXDATASIZE);
		sprintf(buf, "we see %d groups", number_groups);
		//printf("%s\n", buf);
		#ifdef TEST_WIFI
		printf("previous number groups is %d\n", previous_number_groups);
		printf("current number groups is %d\n", number_groups);
		printf("\n");
		delay(500);
		#endif
		if ((previous_number_groups != number_groups) && (numbytes = send(sockfd, buf, MAXDATASIZE-1, 0)) == -1) {
	       		perror("recv");
         		exit(1);
        	}

		previous_number_groups = number_groups;
	   	buf[numbytes] = '\0';
	   	//close(sockfd);

	}

}
